/**************************************************************************************/
/*                                                                                    */
/* File:    CRT0.S                                                                    */
/* Author:  Klemen Kisel                                                              */
/* Date:    20.10.2005                                                                */
/*                                                                                    */
/* Description: LPC21XX crt0 startup file.                                            */
/*                                                                                    */
/**************************************************************************************/


/**************************************************************************************/
/* Configuration                                                                      */
/**************************************************************************************/
                .equ    VPBDIV_VAL,     0x00000000
                .equ    PLLCFG_VAL,     0x00000024
                
/**************************************************************************************/
/* Definitions                                                                        */
/**************************************************************************************/
                .equ    MODE_USR, 0x10
                .equ    MODE_FIQ, 0x11
                .equ    MODE_IRQ, 0x12
                .equ    MODE_SVC, 0x13
                .equ    MODE_ABT, 0x17
                .equ    MODE_UND, 0x1B
                .equ    MODE_SYS, 0x1F

                .equ    I_BIT, 0x80	            /* when I bit is set, IRQ is disabled */
                .equ    F_BIT, 0x40	            /* when F bit is set, FIQ is disabled */
        
                .equ    STACK_TOP, 0x40004000
                .equ    UND_STACK, 0x00000008
                .equ    SVC_STACK, 0x00000008
                .equ    ABT_STACK, 0x00000008
                .equ    FIQ_STACK, 0x00000008
                .equ    IRQ_STACK, 0x00000080
                .equ    USR_STACK, 0x00000400
        
                .equ    FIQ_STACK_ADDRESS, 0
                .equ    IRQ_STACK_ADDRESS, FIQ_STACK_ADDRESS + FIQ_STACK
                .equ    ABT_STACK_ADDRESS, IRQ_STACK_ADDRESS + IRQ_STACK
                .equ    UND_STACK_ADDRESS, ABT_STACK_ADDRESS + ABT_STACK
                .equ    SVC_STACK_ADDRESS, UND_STACK_ADDRESS + UND_STACK
                .equ    USR_STACK_ADDRESS, SVC_STACK_ADDRESS + SVC_STACK

                .equ    VPBDIV,         0xE01FC100

                .equ    PLL_BASE,       0xE01FC080
                .equ    PLLCON_OFS,     0x00
                .equ    PLLCFG_OFS,     0x04
                .equ    PLLSTAT_OFS,    0x08
                .equ    PLLFEED_OFS,    0x0C
                .equ    PLLCON_PLLE,    (1<<0)
                .equ    PLLCON_PLLC,    (1<<1)
                .equ    PLLCFG_MSEL,    (0x1F<<0)
                .equ    PLLCFG_PSEL,    (0x03<<5)
                .equ    PLLSTAT_PLOCK,  (1<<10)
        
/**************************************************************************************/
/* Startup code. First adderss.                                                       */
/**************************************************************************************/

                .text
                .arm
        
                .global _start
                .global main
                .func   _start


_start:

/* Program signature explanation:                                                 */
/* Interrupt vector table is 32 bytes long (eight 32-bit instructions). The sixth */
/* instruction is a 32-bit constant named program signature. It must be a 32-bit  */
/* number such that the eight 32-bit instruction codes from interrupt vector      */
/* table produce zero when they are added together (without carry).               */
/* The signature remains unchanged if the interrupt service routine addresses are */
/* listed in a separate table (interrupt address table).                          */  

/* Interrupt vector table */
                   ldr       pc, reset_addr
                   ldr       pc, undefined_addr
                   ldr       pc, swi_addr
                   ldr       pc, prefetch_addr
                   ldr       pc, abort_addr
                   .long     0xb8a06f58                    /* Program signature */
                   ldr       pc, irq_addr
                   ldr       pc, fiq_addr

/* Interrupt address table */
/* Replace the handler address with C routine name if you want to */
/* use some other function that handles an interrupt.             */
reset_addr:        .long     Reset_Handler
undefined_addr:    .long     undefined_except
swi_addr:          .long     swi_except
prefetch_addr:     .long     prefetch_except
abort_addr:        .long     abort_except
                   .long     0x00000000                    /* Reserved four bytes */
irq_addr:          .long     irq_except
fiq_addr:          .long     fiq_except

/* Default interrupt service routines.      */
/* Hang the system in case of an interrupt. */
undefined_except:  b         undefined_except
swi_except:        b         swi_except
prefetch_except:   b         prefetch_except
abort_except:      b         abort_except
irq_except:        b         irq_except
fiq_except:        b         fiq_except
          

/**************************************************************************************/
/* Reset Handler                                                                      */
/**************************************************************************************/
Reset_Handler:
                /*
                   value = 0 ... VPB clock is 1/4 of the processor clock
                */
                /* VPB Serup */
                LDR     R0, =VPBDIV
                LDR     R1, =VPBDIV_VAL
                STR     R1, [R0]
                
                /* 
                   12MHz XTAL
                */
                /* PLL Setup */
                LDR     R0, =PLL_BASE
                MOV     R1, #0xAA
                MOV     R2, #0x55

                /* PLLCFG_VAL .. 0x24 = 001 00100 
                     M=4 (bits 0..4 = 4)
                     P=2 (bits 5..6 = 1 dec = 01 binary)
                          00 = 1
                          01 = 2
                          10 = 4
                          11 = 8
                           
                   FOSC=12MHz
                   CCLK=(M+1)*FOSC=60MHz       .. processor clock
                   FCCO=CCLK*2*P=240MHz        .. must be between 156MHz and 320MHz
                */
                MOV     R3, #PLLCFG_VAL
                STR     R3, [R0, #PLLCFG_OFS]
                /* Set PLLE bit (bit 0) in PLLCON */ 
                MOV     R3, #PLLCON_PLLE
                STR     R3, [R0, #PLLCON_OFS]
                /* Write feed sequence 0xaa, 0x55 to PLLFEED */
                STR     R1, [R0, #PLLFEED_OFS]
                STR     R2, [R0, #PLLFEED_OFS]

                /* Wait for PLL */
                /* 
                   Check PLOCK bit (bit 10) in PLLSTAT
                   Wait until it becomes 1.
                */
PLL_Loop:       LDR     R3, [R0, #PLLSTAT_OFS]
                ANDS    R3, R3, #PLLSTAT_PLOCK
                BEQ     PLL_Loop

                /* Enable PLL Clock */
                /*
                   PLLE set to 1 keeps PLL enabled
                   PLLC connects PLL as the source of CPU clock
                */
                MOV     R3, #(PLLCON_PLLE | PLLCON_PLLC)
                STR     R3, [R0, #PLLCON_OFS]
                /* Write feed sequence 0xaa, 0x55 to PLLFEED */                
                STR     R1, [R0, #PLLFEED_OFS]
                STR     R2, [R0, #PLLFEED_OFS]

                /* Setup Stack for each mode */
                LDR     R0, =STACK_TOP
                MSR     CPSR_c, #MODE_FIQ|I_BIT|F_BIT        /* Set FIQ stack pointer */
                MOV     SP, R0
                SUB     R0, R0, #FIQ_STACK
                MSR     CPSR_c, #MODE_UND|I_BIT|F_BIT        /* Set UND stack pointer */
                MOV     SP, R0
                SUB     R0, R0, #UND_STACK
                MSR     CPSR_c, #MODE_ABT|I_BIT|F_BIT        /* Set ABT stack pointer */
                MOV     SP, R0
                SUB     R0, R0, #ABT_STACK
                MSR     CPSR_c, #MODE_IRQ|I_BIT|F_BIT        /* Set IRQ stack pointer */
                MOV     SP, R0
                SUB     R0, R0, #IRQ_STACK
                MSR     CPSR_c, #MODE_SVC|I_BIT|F_BIT        /* Set SVC stack pointer */
                MOV     SP, R0
                SUB     R0, R0, #SVC_STACK
                MSR     CPSR_c, #MODE_USR                    /* Set SYS stack pointer */
                MOV     SP, R0

/**************************************************************************************/
/* Copy form ROM to RAM.                                                              */
/**************************************************************************************/
                LDR     R1, =_etext
                LDR     R2, =_data
                LDR     R3, =_edata
LoopRel:        CMP     R2, R3
                LDRLO   R0, [R1], #4
                STRLO   R0, [R2], #4
                BLO     LoopRel


/**************************************************************************************/
/* Clear bss RAM section.                                                             */
/**************************************************************************************/
                MOV     R0, #0
                LDR     R1, =__bss_start__
                LDR     R2, =__bss_end__
LoopZI:         CMP     R1, R2
                STRLO   R0, [R1], #4
                BLO     LoopZI

/**************************************************************************************/
/* Enter C main function.                                                             */
/**************************************************************************************/
                B       main


                .size   _start, . - _start
                .endfunc
        
        
                .end
