#include "io.h"

void LPC21XX_UNDInterruptHandler(void)
{
    while(1)
    {
    }
}

void LPC21XX_SWIInterruptHandler(void)
{
    while(1)
    {
    }                                   
}

void LPC21XX_ABTInterruptHandler(void)
{
    while(1)
    {
    }
}

void LPC21XX_IRQInterruptHandler(void)
{
    while(1)
    {
    }
}

void LPC21XX_FIQInterruptHandler(void)
{
    while(1)
    {
    }
}

//-----------------------------------------------------------------------------
//-- Wrapper functions start here
//-- 
//-----------------------------------------------------------------------------


//-----------------------------------------------------------------------------
//LCD
//-----------------------------------------------------------------------------

char _printfBuffer[100]; //used by printf macro
char _kbdbuffer = 0; //holds the last key pressed
char _LCDBuffer[16]; //holds second line of LCD for the purpose of scrolling
void _LCDInit(void)
{
    // Set up timer1 for LCD delay loop
    SLEEPLOOPINIT;
    
    // Set up CPU for LCD
    LCD_INIT;
    
    // Voodoo LCD init
    LCDVOODOO;
    
    // Command (8-bit off, two-line, big characters off)
    LCDCMD_SETFUNC(0, 0, 1, 0); 

    // Command (turn on display, cursor, and blink)
    LCDCMD_ONOFF(0, 1, 1, 1);
    
    // Command (clear display) 
    LCDCMD_CLEAR(0);

    // Command (entry mode set) ... increment and shift
    LCDCMD_ENTRYMODE(0, 1, 0);

    // Clear display and buffer
    clrscr();
}

void clrscr(void)
//clear display and move cursor to position (1, 1)
{
  LCDCMD_CLEAR(1);
  memset(_LCDBuffer, ' ', 16);
}

void scroll(void)
{
  int i;
  LCDCMD_SETDDRADDR(1, 0);
  for (i = 0; i < 16; i++)
  {
    LCDCMD_WRITEASCII(_LCDBuffer[i]);
  }
  LCDCMD_SETDDRADDR(1, 0x40);
  for (i = 0; i < 16; i++)
  {
    LCDCMD_WRITEASCII(' ');
  }
  LCDCMD_SETDDRADDR(1, 0x40);
  memset(_LCDBuffer, ' ', 16);
}

void putch(int ch)
//write character at the cursor position
//at the end of line 1 move cursor to the beginning of line 2
//at the end of line 2 move cursor to the beginning of line 2 and scroll up
//recognizes newline ('\n') and return ('\r') 
{
  int i;
  long busy = 1, addr;
  //while (busy) LCDCMD_READADDR(busy, addr);  

  if (ch == '\r') //return
  { 
    if (addr < 0x40) 
    {
      LCDCMD_SETDDRADDR(1,0);
    }
    else 
    {
      LCDCMD_SETDDRADDR(1,0x40);
    }
    return;
  }
  else if (ch == '\n') //newline
  {
    if (addr >= 0x40) 
    {
      scroll();
    }
    else
    {
      addr = min(addr + 0x40, 0x67);
    }
    LCDCMD_SETDDRADDR(1, addr);
    return;
  } 
  else if (ch == '\b') //backspace
  {
    if (addr != 0 && addr != 0x40)
    {
      addr -= 1;
    }
    else if (addr == 0x40)
    {
      addr = 0xF;
    }
    
    LCDCMD_SETDDRADDR(1, addr);
    LCDCMD_WRITEASCII(' ');  
    LCDCMD_SETDDRADDR(1, addr);
    return;
  }
  if (ch > '}' || ch < ' ')
  {
    return;
  }
  
  LCDCMD_WRITEASCII(ch);
  busy = 1;
  while (busy) LCDCMD_READADDR(busy, addr);  
  if (addr > 0x0F && addr < 0x40) 
  {
    LCDCMD_SETDDRADDR(1, 0x40);
  }
  if (addr > 0x40 && addr <= 0x50)
  {
    _LCDBuffer[addr - 1 - 0x40] = ch;    
  }     
  if (addr > 0x4F) 
  {
    scroll();
  }
}

void _setcursortype(int cur_t)
//control the appearence of curcor
{
  switch(cur_t)
  {
  case _NOCURSOR: LCDCMD_ONOFF(1, 1, 0, 0); break;
  case _NORMALCURSOR: LCDCMD_ONOFF(1, 1, 1, 0); break;
  case _SOLIDCURSOR: LCDCMD_ONOFF(1, 1, 1, 1); break;
  }
}

void gotoxy(int x, int y)
//position cursor on 2 x 16 LCD
//if given position is invalid, nothing happens
{
  int addr;  
  //if invalid position
  if (x < 1 || y < 1) return;
  if (x > 16 || y > 2) return;
  
  addr = 0x40 * (y - 1) + (x - 1);
  LCDCMD_SETDDRADDR(1, addr);
}

int puts(char *str)
//unlike ANSI C puts(), this version does not output carriage return
{
  int i;
  for (i = 0; str[i] != 0; i++)
  {
    putch(str[i]);
  }
  return i;
}


//-----------------------------------------------------------------------------
//Keys
//-----------------------------------------------------------------------------

void _KeyInit(void)
{
  KEY_INIT;
  _kbdbuffer = 0;
}

int _getkeys(void)
//get state of all four keys
{
  int i, state, revstate = 0;
  KEY_READALL(state, 10, 2000);
  for (i = 0; i < 4; i++)
  {
    revstate |= state % 2;
    state >>= 1;
    revstate <<= 1;    
  }
  return revstate >> 1;
}

int kbhit(void)
//Detektira prehod iz nule v enko med dvema klicema.
//e prehoda ni bilo in e je v spremenljivki _kbdbuffer nula, vrne 0, sicer 
//vrne celotevilsko vrednost, razlino od ni.
//v globalno spremenljivko _kbdbuffer shrani kodo pritisnjene tipke
//
{
  int i;
  static int keys = 0;
  
  int keys1 = _getkeys();
  for (i = 8; i > 0; i >>= 1)
  {
    if ((keys1 & i) > (keys & i)) break;
  }
  keys = keys1;
  switch (i)
  {
  case 8: _kbdbuffer = '0'; break;    
  case 4: _kbdbuffer = '1'; break;    
  case 2: _kbdbuffer = '2'; break;    
  case 1: _kbdbuffer = '3'; break;   
  }
  
  return _kbdbuffer;
}

int getch(void)
{
  char ret;
  while (!kbhit());
  ret = _kbdbuffer;
  _kbdbuffer = 0;
  return ret;
}


//-----------------------------------------------------------------------------
//LEDs
//-----------------------------------------------------------------------------

void _LEDInit(void)
{
  LED_INIT;
}

void _setleds(int state)
//turn on specified LEDs
{
  LED_SET(state);
}

void _clrleds(int state)
//turn off specified LEDs
{
  LED_CLR(state);
}

//-----------------------------------------------------------------------------
//Timer
//-----------------------------------------------------------------------------

void _TimerInit(void)
{
  T0PR 		= 0x00003a97;			// Load prescaler (0x3a97 = 14.999, pr+1=15.000)
  T0MR0		= 0xffffffff;   	// Set the cycle time (ko pride T0TC do te vrednosti, se T0TC resetira)
  T0TCR 	= 0x00000002;			// Reset counter and prescaler
  T0CTCR  &=0xfffffffc;     // Timer mode 
  T0TCR 	= 0x00000001;			// Enable timer
}

unsigned long clock(void)
//return number of milliseconds since CPU start 
{
  return T0TC;
}

void delay(unsigned milliseconds)
//suspend execution for specified time  
{
  unsigned long start = T0TC;
  while (T0TC - start < milliseconds);
//  SLEEPLOOP(29000 * milliseconds);    
}

//-----------------------------------------------------------------------------
//ADC
//-----------------------------------------------------------------------------

void _ADCInit(void)
{
  ADC_INIT;
}

int _adconvert(void)
//convert and return value from ADC
//returned value lies between 0 and 1023 inclusive
{
  ADC_CONVERT;
  return ADC_READ;
}

int _adconvertExt(int input)
//input: 0 selects AD0.0, 7 selects AD0.7
//Needs no initialization
//Notice: _adconvertExt(0); is equivalent to _ADCInit(); _adconvert();
{
  int val = 0x400000 << (input * 2);
  int mask = 0xc00000 << (input * 2);
  int selectpin = 0x1 << input;
  int retval;
  
  if (input < 0 || input > 7) return;

  PINSEL1 = (PINSEL1 & ~mask) | val;

    ADC_CR(AD0) = selectpin | (0x4 << 8) | (0x1 << 21) | (0x1 << 24);
    do { 
      retval = ADC_DR(AD0);
    } while ((retval & (0x1 << 31)) == 0);
    retval = ((ADC_DR(AD0) & 0xffc0) >> 6);

    return retval;
}


//-----------------------------------------------------------------------------
//DAC
//-----------------------------------------------------------------------------

void _DACInit(void)
{
  DAC_INIT;
}

void _dacwrite(int val)
{
  if (val < 0) val = 0;
  if (val > 1023) val = 1023;
  DAC_WRITE(val, 0);
}

//-----------------------------------------------------------------------------
//GPIO
//-----------------------------------------------------------------------------

void _setpindir(unsigned int pinid, int dir)
//pinid je med 0 in 15 in ustreza pinom P0.0 do P0.15
//dir = 0: vhod, dir = 1, izhod
{
  unsigned int tmp;
  unsigned int dirmask;
  unsigned int selectpinmask;
  
  if (pinid > 15) return;

  dirmask = 0x1 << pinid;
  selectpinmask = ~(0x3 << (pinid * 2));

  PINSEL0 &= selectpinmask; //pin postane GPIO port  

  if (dir == 0)
  {
    IO0DIR = IO0DIR & ~dirmask;
  }
  else
  {
    IO0DIR = IO0DIR | dirmask;
  } 

}

void outportp(unsigned int pinid, int value)
//Pie na pin pinid, e je le-ta konfiguriran kot izhod
{
  if (pinid > 15) return;
  if (value == 0)
  {
    IO0CLR = 0x1 << pinid;
  } 
  else
  {
    IO0SET = 0x1 << pinid;
  } 
}

int inportp(unsigned int pinid)
{
  if (pinid > 15) return -1;
  return (IO0PIN & (0x1 << pinid)) >> pinid;
}

