void LPC21XX_UNDInterruptHandler(void) __attribute__((interrupt("IRQ")));
void LPC21XX_SWIInterruptHandler(void) __attribute__((interrupt("IRQ")));
void LPC21XX_ABTInterruptHandler(void) __attribute__((interrupt("IRQ")));
void LPC21XX_IRQInterruptHandler(void) __attribute__((interrupt("IRQ")));
void LPC21XX_FIQInterruptHandler(void) __attribute__((interrupt("FIQ")));


//-----------------------------------------------------------------------------
//-- Hardware drivers
//--   (c)2005 A. Buermen
//-----------------------------------------------------------------------------
//timer 0
#define T0IR           (*((volatile unsigned long *) 0xE0004000))
#define T0TCR          (*((volatile unsigned long *) 0xE0004004))
#define T0TC           (*((volatile unsigned long *) 0xE0004008))
#define T0PR           (*((volatile unsigned long *) 0xE000400C))
#define T0PC           (*((volatile unsigned long *) 0xE0004010))
#define T0MCR          (*((volatile unsigned long *) 0xE0004014))
#define T0MR0          (*((volatile unsigned long *) 0xE0004018))
#define T0MR1          (*((volatile unsigned long *) 0xE000401C))
#define T0MR2          (*((volatile unsigned long *) 0xE0004020))
#define T0MR3          (*((volatile unsigned long *) 0xE0004024))
#define T0CCR          (*((volatile unsigned long *) 0xE0004028))
#define T0CR0          (*((volatile unsigned long *) 0xE000402C))
#define T0CR1          (*((volatile unsigned long *) 0xE0004030))
#define T0CR2          (*((volatile unsigned long *) 0xE0004034))
#define T0CR3          (*((volatile unsigned long *) 0xE0004038))
#define T0EMR          (*((volatile unsigned long *) 0xE000403C))
#define T0CTCR         (*((volatile unsigned long *) 0xE0004070))

//timer 1
#define T1IR           (*((volatile unsigned long *) 0xE0008000))
#define T1TCR          (*((volatile unsigned long *) 0xE0008004))
#define T1TC           (*((volatile unsigned long *) 0xE0008008))
#define T1PR           (*((volatile unsigned long *) 0xE000800C))
#define T1PC           (*((volatile unsigned long *) 0xE0008010))
#define T1MCR          (*((volatile unsigned long *) 0xE0008014))
#define T1MR0          (*((volatile unsigned long *) 0xE0008018))
#define T1MR1          (*((volatile unsigned long *) 0xE000801C))
#define T1MR2          (*((volatile unsigned long *) 0xE0008020))
#define T1MR3          (*((volatile unsigned long *) 0xE0008024))
#define T1CCR          (*((volatile unsigned long *) 0xE0008028))
#define T1CR0          (*((volatile unsigned long *) 0xE000802C))
#define T1CR1          (*((volatile unsigned long *) 0xE0008030))
#define T1CR2          (*((volatile unsigned long *) 0xE0008034))
#define T1CR3          (*((volatile unsigned long *) 0xE0008038))
#define T1EMR          (*((volatile unsigned long *) 0xE000803C))
#define T1CTCR         (*((volatile unsigned long *) 0xE0008070))


// CPU clock
#define CPU_CLOCK_MHZ 60
#define VPB_CLOCK_MHZ 15

// p0.0 .. p0.31 control
#define PINSEL0 (*((unsigned int *)0xe002c000)) 
#define PINSEL1 (*((unsigned int *)0xe002c004)) 
#define IO0DIR  (*((unsigned int *)0xe0028008)) 
#define IO0SET  (*((unsigned int *)0xe0028004)) 
#define IO0CLR  (*((unsigned int *)0xe002800c)) 
#define IO0PIN  (*((unsigned int *)0xe0028000))

// Key pin mask in PINSEL0 and IO0DIR
#define KEYPINSELMASK 0xff000000
#define KEYPINMASK 0xf000

#define KEY_INIT { \
   PINSEL0=PINSEL0 & ~KEYPINSELMASK; \
   IO0DIR=IO0DIR & ~KEYPINMASK; \
}

#define KEY_READALL(result, debounce, max) { \
   int i=0, last=-1, same=0; \
   do { \
     (result)=(~(IO0PIN) & KEYPINMASK); \
     if ((result)==last) { \
       same++; \
     } else { \
       same=0; \
     } \
     last=(result); \
     i++; \
   } while (i<(max) && same<(debounce)); \
   if (i==(max)) { \
     (result)=-1; \
   } else { \
     (result)=(result)>>12; \
   } \
}

// LED pin mask in PINSEL0 and IO0DIR
#define LEDPINSELMASK 0x0000ff00
#define LEDPINMASK 0xf0

#define LED_INIT { \
   PINSEL0=PINSEL0 & ~LEDPINSELMASK; \
   IO0DIR=IO0DIR | LEDPINMASK; \
}

#define LED_SET(mask) { \
   IO0SET = (((mask) & 0xf)<<4); \
}

#define LED_CLR(mask) { \
   IO0CLR = (((mask) & 0xf)<<4); \
}

#define LED_STATE (((IO0PIN)>>4) & 0xf)


// LCD pin mask in PINSEL1 and IO0DIR
#define LCDPINSELMASK 0x3fff
#define LCDPINMASK 0x7f0000
#define LCDDATAPINMASK 0xf0000

// LCD pin masks in IO0SET and IO0CLR
#define LCDENCRS    (0x1<<20)
#define LCDENCRW    (0x1<<22)
#define LCDENCE     (0x1<<21)
#define LCDENCDS(d) ((d)<<16)
#define LCDENCDC(d) ((~((d)<<16))&(0xf<<16))

// Set up timer1 so that it counts the VPB clock tics
#define SLEEPLOOPINIT { \
  T1PR 		= 0x00000000; \
  T1MR0		= 0xffffffff; \
  T1MCR   = 0x00000000; \
  T1IR    = 0xffffffff; \
  T1TCR 	= 0x00000002; \
  T1CTCR  &=0xfffffffc; \
  T1TCR 	= 0x00000001; \
}

// Sleep for a given number of nanoseconds
// Minimal speep (0) equals overhead + 1 tick
// 1 tick = 1/fvpb 
#define SLEEPLOOP(ns) { \
  unsigned int from=T1TC; \
  unsigned int tics=ns*VPB_CLOCK_MHZ/1000+1; \
  while (T1TC-from<tics); \
}

// LCD communication
// Write cycle
#define WRSETUP(rs) { \
   if (rs) \
     IO0SET=LCDENCRS; \
   else \
     IO0CLR=LCDENCRS; \
   IO0CLR=LCDENCRW; \
   SLEEPLOOP(140+20); \
}

#define WRDATA(data) { \
   IO0SET=LCDENCE; \
   IO0SET=LCDENCDS((data)); \
   IO0CLR=LCDENCDC((data)); \
   SLEEPLOOP(450+20); \
}

#define WREND(e) { \
   IO0CLR=LCDENCE; \
   SLEEPLOOP(20+5); \
   IO0SET=LCDENCRW; \
   SLEEPLOOP(600+20); \
   if (e) \
     IO0SET=LCDENCE; \
   else \
     IO0CLR=LCDENCE; \
}

// Read cycle
#define RDSETUP(rs) { \
   if (rs) \
     IO0SET=LCDENCRS; \
   else \
     IO0CLR=LCDENCRS; \
   IO0DIR=IO0DIR & ~LCDDATAPINMASK; \
   IO0SET=LCDENCRW; \
   SLEEPLOOP(140+20); \
   IO0SET=LCDENCE; \
   SLEEPLOOP(320+20); \
}


#define RDDATA ((IO0PIN & LCDDATAPINMASK) >> 16)

#define RDEND(e) { \
   SLEEPLOOP(150+20); \
   IO0CLR=LCDENCE; \
   SLEEPLOOP(20+5); \
   IO0CLR=LCDENCRW; \
   IO0DIR=IO0DIR | LCDDATAPINMASK; \
   SLEEPLOOP(600+20); \
   if (e) \
     IO0SET=LCDENCE; \
   else \
     IO0CLR=LCDENCE; \
}

// Voodoo with LCD before the real stuff starts
// This is partially fropm page 46 of HITACHI datasheet
// for HD44780U (LCD-II). These strange things have to 
// be done before the function set command that switches 
// to 4-bit mode is issued.
// The sequence that follows is:
//   - function set
//   - display on/off control
//   - display clear
//   - entry mode set
// All commands must be issued without polling the busy flag. 
#define LCDVOODOO { \
  SLEEPLOOP(50000000); \
  WRSETUP(0); \
  WRDATA(0x3); \
  WREND(0); \
  SLEEPLOOP(4500000); \
  WRSETUP(0); \
  WRDATA(0x3); \
  WREND(0); \
  SLEEPLOOP(200000); \
  WRSETUP(0); \
  WRDATA(0x3); \
  WREND(0); \
  WRSETUP(0); \
  WRDATA(0x2); \
  WREND(0); \
}

// Waiting for busy to go low
#define LCDWAITBUSY(busy) { \
   do { \
     RDSETUP(0); \
     (busy)=RDDATA & 0x8; \
     RDEND(0); \
     RDSETUP(0); \
     RDEND(0); \
   } while (busy); \
}

// Basic LCD commands
#define LCD_INIT { \
   PINSEL1=PINSEL1 & ~LCDPINSELMASK; \
   IO0DIR=IO0DIR | LCDPINMASK; \
}

#define LCDCMD_CLEAR(dowait) { \
   WRSETUP(0); \
   WRDATA(0x0); \
   WREND(0); \
   \
   WRSETUP(0); \
   WRDATA(0x1); \
   WREND(0); \
   \
   if (dowait) { \
     int busy; \
     LCDWAITBUSY(busy); \
   } \
}

#define LCDCMD_HOME(dowait) { \
   WRSETUP(0); \
   WRDATA(0x0); \
   WREND(0); \
   \
   WRSETUP(0); \
   WRDATA(0x2); \
   WREND(0); \
   \
   if (dowait) { \
     int busy; \
     LCDWAITBUSY(busy); \
   } \
}

#define LCDCMD_ENTRYMODE(dowait, increment, shift) { \
   WRSETUP(0); \
   WRDATA(0x0); \
   WREND(0); \
   \
   WRSETUP(0); \
   WRDATA((0x4 | ((increment)<<1) | (shift))); \
   WREND(0); \
   \
   if (dowait) { \
     int busy; \
     LCDWAITBUSY(busy); \
   } \
}

#define LCDCMD_ONOFF(dowait, display, cursor, blink) { \
   WRSETUP(0); \
   WRDATA(0x0); \
   WREND(0); \
   \
   WRSETUP(0); \
   WRDATA((0x8 | ((display)<<2) | ((cursor)<<1) | (blink))); \
   WREND(0); \
   \
   if (dowait) { \
     int busy; \
     LCDWAITBUSY(busy); \
   } \
}

#define LCDCMD_SHIFT(dowait, displayshift, shiftright) { \
   WRSETUP(0); \
   WRDATA(0x1); \
   WREND(0); \
   \
   WRSETUP(0); \
   WRDATA((((displayshift)<<3) | ((shiftright)<<2))); \
   WREND(0); \
   \
   if (dowait) { \
     int busy; \
     LCDWAITBUSY(busy); \
   } \
}

#define LCDCMD_SETFUNC(dowait, eightbit, twoline, bigchar) { \
   WRSETUP(0); \
   WRDATA((0x2 | (eightbit))); \
   WREND(0); \
   \
   WRSETUP(0); \
   WRDATA((((twoline)<<3) | ((bigchar)<<2))); \
   WREND(0); \
   \
   if (dowait) { \
     int busy; \
     LCDWAITBUSY(busy); \
   } \
}

#define LCDCMD_SETDDRADDR(dowait, addr) { \
   WRSETUP(0); \
   WRDATA(0x8 | (((addr) & 0x70)>>4)); \
   WREND(0); \
   \
   WRSETUP(0); \
   WRDATA((addr) & 0xf); \
   WREND(0); \
   \
   if (dowait) { \
     int busy; \
     LCDWAITBUSY(busy); \
   } \
}

#define LCDCMD_SETCGRADDR(dowait, addr) { \
   WRSETUP(0); \
   WRDATA(0x4 | (((addr) & 0x30)>>4)); \
   WREND(0); \
   \
   WRSETUP(0); \
   WRDATA((addr) & 0xf); \
   WREND(0); \
   \
   if (dowait) { \
     int busy; \
     LCDWAITBUSY(busy); \
   } \
}

#define LCDCMD_WRITE(dowait, data) { \
   WRSETUP(1); \
   WRDATA((((data) & 0xf0)>>4)); \
   WREND(0); \
   \
   WRSETUP(1); \
   WRDATA(((data) & 0xf)); \
   WREND(0); \
   \
   if (dowait) { \
     int busy; \
     LCDWAITBUSY(busy); \
   } \
}

#define LCDCMD_READADDR(busy, addr) { \
   (busy)=0; \
   (addr)=0; \
   RDSETUP(0); \
   (addr)=RDDATA; \
   RDEND(0); \
   (busy) = ((addr) & 0x8) >> 3; \
   (addr) = ((addr) & 0x7) << 4; \
   RDSETUP(0); \
   (addr) = (addr) | (RDDATA & 0xf); \
   RDEND(0); \
}

#define LCDCMD_READDATA(data) { \
   RDSETUP(1); \
   (data)=RDDATA; \
   RDEND(0); \
   (data) = ((data) & 0xf) << 4; \
   RDSETUP(1); \
   (data) = (data) | (RDDATA & 0xf); \
   RDEND(0); \
}

// Advanced LCD commands
#define LCDCMD_WRITEASCII(ascii) { \
   if ((ascii)>=0x20 && (ascii)<=0x7f) { \
     LCDCMD_WRITE(1, (ascii)); \
   } else { \
     LCDCMD_WRITE(1, 0x20); \
   } \
}

// Advances LCD functions
void LCDCMD_write(char *str);

// UART control
#define UART0PINSELMASK 0xf
#define UART0PINSELVAL  0x5

#define UART1PINSELMASK 0xf0000
#define UART1PINSELVAL  0x50000

#define UART0 0xe000c000
#define UART1 0xe0010000

#define UART_RBR(base) (*((unsigned int *)((base)+0x0))) 
#define UART_THR(base) (*((unsigned int *)((base)+0x0))) 
#define UART_DLL(base) (*((unsigned int *)((base)+0x0))) 
#define UART_DLM(base) (*((unsigned int *)((base)+0x4))) 
#define UART_IER(base) (*((unsigned int *)((base)+0x4))) 
#define UART_IIR(base) (*((unsigned int *)((base)+0x8))) 
#define UART_FCR(base) (*((unsigned int *)((base)+0x8))) 
#define UART_LCR(base) (*((unsigned int *)((base)+0xc))) 
#define UART_MCR(base) (*((unsigned int *)((base)+0x10))) 
#define UART_LSR(base) (*((unsigned int *)((base)+0x14))) 
#define UART_MSR(base) (*((unsigned int *)((base)+0x18))) 
#define UART_SCR(base) (*((unsigned int *)((base)+0x1c)))

#define UART_5bit 0x0
#define UART_6bit 0x1
#define UART_7bit 0x2
#define UART_8bit 0x3

#define UART_1stop 0x0
#define UART_2stop 0x4

#define UART_noparity 0x0
#define UART_oddparity (0x8 | (0x0<<4))
#define UART_evenparity (0x8 | (0x1<<4)) 
#define UART_oneparity (0x8 | (0x2<<4)) 
#define UART_zeroparity (0x8 | (0x3<<4))

#define UART_INIT(base) { \
   if ((base)==UART0) { \
     PINSEL0=(PINSEL0 & ~UART0PINSELMASK) | UART0PINSELVAL; \
   } else { \
     PINSEL0=(PINSEL0 & ~UART1PINSELMASK) | UART1PINSELVAL; \
   } \
}

#define UART_SETUP(base, vpb_mhz, baudrate, wordlength, stop, parity) { \
   unsigned int divisor=(unsigned int)(1000000*vpb_mhz/(16*baudrate))-1; \
   UART_LCR(base)=UART_LCR(base)|0x80; \
   UART_DLL(base)=(divisor) & 0xff; \
   UART_DLM(base)=((divisor) & 0xff00)>>8; \
   UART_LCR(base)=(wordlength) | (stop) | (parity); \
   }

#define UART_SEND(base, data) { \
   UART_THR(base)=(data); \
}

#define UART_DATA_SENT(base) (UART_LSR(base) & 0x20)

#define UART_READ(base) UART_RBR(base)

#define UART_DATA_READY(base) (UART_LSR(base) & 0x01)


#define DAC_PINSELMASK 0xc0000
#define DAC_PINSELVAL  0x80000

#define DAC_REGISTER (*((unsigned int *)(0xe006c000)))

#define DAC_INIT { \
   PINSEL1 = (PINSEL1 & ~DAC_PINSELMASK) | DAC_PINSELVAL; \
   }

#define DAC_WRITE(value, bias) { \
   DAC_REGISTER = (((bias) & 0x1) << 16) | (((value) & 0x3ff) << 6); \
   }


// AD0.0
#define ADC_PINSELMASK 0xc00000
#define ADC_PINSELVAL  0x400000

#define AD0 0xe0034000
#define AD1 0xe0060000

#define ADC_CR(base) (*((unsigned int *)(base))) 
#define ADC_DR(base) (*((unsigned int *)(base+0x4))) 
#define ADC_GSR (*((unsigned int *)(0xe0034008)))

// Divide VPB clock by 4 to get the ADC clock, burst=0 
#define ADC_INIT { \
   PINSEL1 = (PINSEL1 & ~ADC_PINSELMASK) | ADC_PINSELVAL; \
   ADC_CR(AD0) = 0x1 | (0x4 << 8) | (0x1 << 21); \
   }

#define ADC_CONVERT { \
   unsigned int tmp; \
   ADC_CR(AD0) = 0x1 | (0x4 << 8) | (0x1 << 21) | (0x1 << 24); \
   do { \
     tmp = ADC_DR(AD0); \
   } while ((tmp & (0x1 << 31)) == 0); \
   }

#define ADC_READ ((ADC_DR(AD0) & 0xffc0) >> 6)


//-----------------------------------------------------------------------------
//-- Hardware drivers end here
//--
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
//-- Wrapper functions start here
//-- 
//-----------------------------------------------------------------------------

#define min(x, y) (x) > (y) ? y : x
#define _NOCURSOR      0
#define _SOLIDCURSOR   1
#define _NORMALCURSOR  2

extern char _LCDBuffer[16]; //holds second line of LCD for the purpose of scrolling
void _LCDInit(void);
void clrscr(void);
void scroll(void);
void putch(int ch);
void _setcursortype(int cur_t);
void gotoxy(int x, int y);
void _KeyInit(void);
int _getkeys(void);
void _LEDInit(void);
void _setleds(int state);
void _clrleds(int state);
void _TimerInit(void);
unsigned long clock(void);
void delay(unsigned milliseconds);
void _ADCInit(void);
int _adconvert(void);
int _adconvertExt(int input);
void _DACInit(void);
void _dacwrite(int val);
void _setpindir(unsigned int pinid, int dir);
void outportp(unsigned int pinid, int value);
int inportp(unsigned int pinid);

extern char _kbdbuffer; //holds the last key pressed

int puts(char *str);
int getch(void);
int kbhit(void);

extern char _printfBuffer[100];
#define printf(...) snprintf(_printfBuffer, 99, __VA_ARGS__),puts(_printfBuffer) 

